// MQL4&5-code

// Web-    .
// #define HISTORYTICKS_RANNFOREX //       .

#property strict

#property indicator_chart_window
#property indicator_buffers 0
#property indicator_plots indicator_buffers

sinput bool AllCharts = true;  // true - All Charts, false - Current Chart
sinput bool inSelfEvent = false; // Send Ticks to the Base Chart (true) or not (false)
sinput int InitHoursTicks = 24 * 7;
sinput uint AmountTicksToSave = 100; // AmountTicksToSave (0 - without save)

#include <fxsaber\HistoryTicks\ArrayResize.mqh>
#include <fxsaber\HistoryTicks\Data_String.mqh>
#include <fxsaber\HistoryTicks\Array.mqh>
#include <fxsaber\HistoryTicks\ResourceData.mqh>

const string Name = "HistoryTicks_" + _Symbol;
string sFileName; //    OnInit.

MqlTick HistoryTicks[];
int AmountTicks = 0;
int AmountCopyTicks = 0;

MqlTick PrevTick = {0};

class WEB
{
public:
  static bool Get( uchar &Data[], const string URL, const int TimeOut = 5000 )
  {
    string Headers;
    char Post[];

    ::Print("Downloading " + URL);

    ulong StartTime = ::GetTickCount();

    const bool Res = (::WebRequest("GET", URL, NULL, NULL, TimeOut, Post, 0, Data, Headers) != -1);

    if (Res)
    {
      StartTime = ::GetTickCount() - StartTime;

      ::Print("Done! - " + (string)::ArraySize(Data) + " bytes" + (StartTime ? " (" + (string)(::ArraySize(Data) / StartTime) + " Kb/s.)" : NULL));
    }
    else
    {
      const int Error = ::GetLastError();
      ::Print("WebRequest Error - " + (string)Error);

      if (Error == ERR_FUNCTION_NOT_CONFIRMED)
        ::Print("Please add \"" + URL + "\" domain name to the list of allowed URLs!");
    }

    return(Res);
  }

  static bool IsURL( const string Str )
  {
    return(::StringFind(Str, "://") != -1);
  }
};

#define ALERT(A) Alert(MQLInfoString(MQL_PROGRAM_NAME) + ": " + __FUNCSIG__ + ", Line = " + (string)__LINE__ + ": " + #A + " - Error!")

string TickToString( const MqlTick &Tick, const ushort Delimeter = '\t', const int digits = 6 )
{
  return((string)Tick.time + ShortToString(Delimeter) + DoubleToString(Tick.bid, digits) +
                             ShortToString(Delimeter) + DoubleToString(Tick.ask, digits));
}

int TicksToFile( const MqlTick &Ticks[], const string FileName, const bool Rewrite = false )
{
  const int Size = ArraySize(Ticks);

  if (Size)
  {
    const int handle = FileOpen(FileName, FILE_WRITE | FILE_ANSI | FILE_TXT | (Rewrite ? 0 : FILE_READ));

    if (handle != INVALID_HANDLE)
    {
      FileSeek(handle, 0, SEEK_END);

      for (int i = 0; i < Size; i++)
        if (!FileWrite(handle, TickToString(Ticks[i])))
        {
          ALERT(FileWrite(handle, TickToString(Ticks[i])));

          break;
        }

      FileClose(handle);
    }
  }

  return(Size);
}

datetime StringToTime( const uchar &Bytes[], const uint &Pos )
{
  static MqlDateTime PrevDate = {0};
  static datetime PrevTime = 0;

  static MqlDateTime DateTime = {0};

  DateTime.year = Bytes[Pos + 0] * 1000 + Bytes[Pos + 1] * 100 + Bytes[Pos + 2] * 10 + Bytes[Pos + 3] - '0' * 1111;
  DateTime.mon = Bytes[Pos + 5] * 10 + Bytes[Pos + 6] - '0' * 11;
  DateTime.day = Bytes[Pos + 8] * 10 + Bytes[Pos + 9] - '0' * 11;

  if ((PrevDate.day != DateTime.day) || (PrevDate.mon != DateTime.mon) || (PrevDate.year != DateTime.year))
  {
    PrevTime = ::StructToTime(DateTime);

    PrevDate = DateTime;
  }

  return(PrevTime +
         (Bytes[Pos + 11] * 10 + Bytes[Pos + 12] - '0' * 11) * 3600 +
         (Bytes[Pos + 14] * 10 + Bytes[Pos + 15] - '0' * 11) * 60 +
          Bytes[Pos + 17] * 10 + Bytes[Pos + 18] - '0' * 11);
}

// 2019.07.22 12:14:49	1.101080	1.101090
// 2019-08-05 00:05:02.685,1.21344,1.21522

double StringToDouble( const uchar &Bytes[], uint &Pos )
{
  static const double TenPow[] = {1, 10, 100, 1000, 10000, 100000, 1000000};

  const uint Size = ::ArraySize(Bytes);

  ulong Res = 0;
  uint point = 0;

  while (Pos < Size)
  {
    const uchar Digit = Bytes[Pos];

    if (((Digit >= '0') && (Digit <= '9')))
      Res = Res * 10 + Digit - '0';
    else if (Digit == '.')
      point = Pos + 2;
    else
    {
      Pos++;

      break;
    }

    Pos++;
  }

  return(Res / TenPow[Pos - point]);
}

bool StringToTick( const uchar &Bytes[], MqlTick &Tick, uint &Pos, const int Step = 20 )
{
  Tick.time = StringToTime(Bytes, Pos);
  Tick.time_msc = Tick.time * 1000;

  Pos += Step;

  Tick.bid = StringToDouble(Bytes, Pos);
  Tick.ask = StringToDouble(Bytes, Pos);

  return(true);
}

int StringToTicks( const uchar &Bytes[], MqlTick &Ticks[], const int Step = 20 )
{
  const uint Size = ::ArraySize(Bytes);
  bool FirstTick = true;
  int Amount = 0;

  MqlTick Tick = {0};

  for (uint i = 0; i < Size; i++)
  {
    StringToTick(Bytes, Tick, i, Step);

    if (FirstTick)
    {
      ArrayResize(Ticks, (int)(Size / (i + 1)));

      FirstTick = false;
    }

    Ticks[Amount++] = Tick;
  }

  return(ArrayResize(Ticks, Amount));
}

int FileToTicks( const string FileName, MqlTick &Ticks[], const int Step = 20 )
{
  uchar Bytes[];

  return((FileLoad(FileName, Bytes) > 0) ? StringToTicks(Bytes, Ticks, Step) : 0);
}

int WebToTicks( const string URL, MqlTick &Ticks[], const int Step = 24 )
{
  uchar Bytes[];

  return(WEB::Get(Bytes, URL) ? StringToTicks(Bytes, Ticks, Step) : 0);
}

int GetPosTick( const MqlTick &Ticks[], const datetime SearchTime )
{
  int Res = 0;

  for (int i = ArraySize(Ticks) - 1; i >= 0; i--)
    if (Ticks[i].time < SearchTime)
    {
      Res = i + 1;

      break;
    }

  return(Res);
}

datetime GetTime( const datetime Offset )
{
  static const ENUM_TIMEFRAMES Periods[] = {PERIOD_CURRENT, PERIOD_M1, PERIOD_M5, PERIOD_M15, PERIOD_M30, PERIOD_H1, PERIOD_H4, PERIOD_D1};

  datetime Res = 0;

  for (int i = 0; !Res && (i < ArraySize(Periods)); i++)
    Res = iTime(_Symbol, Periods[i], (int)Offset / PeriodSeconds(Periods[i]));

  if (!Res)
    ALERT(Res);

  return(Res);
}

#define FILENAME (__FILE__ + ".tpl")
#define PATH "\\Files\\"

// https://www.mql5.com/ru/docs/files/fileload
long FileLoad( const string FileName, uchar &Buffer[], const int CommonFlag = 0 )
{
  long Res = -1;

  const int handle = FileOpen(FileName, FILE_READ | FILE_BIN | CommonFlag);

  if (handle != INVALID_HANDLE)
  {
    if (!(Res = FileReadArray(handle, Buffer)))
    {
      ALERT(FileReadArray(handle, Buffer));

      Res = -1;
    }

    FileClose(handle);
  }

  return(Res);
}

string TemplateToString( const long Chart_ID = 0 )
{
  uchar Data[];

  return((::ChartSaveTemplate((ulong)Chart_ID, PATH + FILENAME) && (::FileLoad(FILENAME, Data) > 0)) ? ::CharArrayToString(Data) : NULL);
}

string GetMyName( void )
{
  const int Len1 = StringLen(TerminalInfoString(TERMINAL_DATA_PATH));
  const int Len2 = StringLen(MQLInfoString(MQL_PROGRAM_PATH));

  return(StringSubstr(MQLInfoString(MQL_PROGRAM_PATH), Len1 + 17, Len2 - Len1 - 21));
}

int IsRun2( const long Chart_ID = 0, const int Pos = 0 )
{
  return(StringFind(TemplateToString(Chart_ID), "<indicator>\r\nname=Custom Indicator\r\n<expert>\r\nname=" + GetMyName(), Pos));
}

bool IsRun( void )
{
  const long Chart_ID = ChartID();

  bool Res = (GetMyName() != "HistoryTicks");

  if (Res)
    Alert("Rename/move Indicator to Indicators\\HistoryTicks.ex4");
  else
  {
    for (long chartID = ChartFirst(); (chartID != -1) && !Res; chartID = ChartNext(chartID))
      if (chartID != Chart_ID)
        Res = (ChartSymbol(chartID) == _Symbol) && (IsRun2(chartID) != -1);
      else
      {
        const int Find = IsRun2(chartID);

        Res = (Find != -1) && (IsRun2(chartID, Find + 1) != -1);
      }

    FileDelete(FILENAME);

    if (Res)
      Alert(Name + " indicator can not start - copy already works.");
  }

  return(Res);
}

#undef PATH
#undef FILENAME

int OnInit()
{
  bool Res = IsRun();

  if (!Res)
  {
    if (Res  = !EventSetTimer(1))
      ALERT(EventSetTimer(1));
    else
      OnTimer();
  }

  return(Res);
}

bool IsReadyToInit()
{
  static string PrevAccountServer = "";

  const string AccountServer = AccountInfoString(ACCOUNT_SERVER);
  const bool Res = (PrevAccountServer != AccountServer);

  if (Res)
    PrevAccountServer = AccountServer;

  return(Res);
}

// https://www.mql5.com/ru/forum/322209/page3#comment_13718775
bool IsOnInit()
{
  static bool Init = false;

  if (!Init)
  {
    if (Init = IsReadyToInit())
    {
      OnDeinit(REASON_ACCOUNT);

      MyOnInit();
    }
  }
  else if (IsReadyToInit())
  {
    OnDeinit(REASON_ACCOUNT);

    MyOnInit();
  }

  return(Init);
}

void CorrectFileName( string &FileName )
{
  while (!IsStopped() && (StringReplace(FileName, "..", ".") > 0))
    ;

  return;
}

void MyOnInit()
{
  GlobalVariableDel(Name);

  AmountTicks = 0;
  AmountCopyTicks = 0;
  ArrayFree(HistoryTicks);

  MqlTick LoadTicks[];

  sFileName = "HistoryTicks\\" + AccountInfoString(ACCOUNT_SERVER) + "\\" + _Symbol + ".csv";

  CorrectFileName(sFileName);

  int AmountLoadTicks = (InitHoursTicks > 0) ? FileToTicks(sFileName, LoadTicks) : 0;
  int AmountWebTicks = 0;

#ifdef HISTORYTICKS_RANNFOREX
  MqlTick WebTicks[];

  //  CSV      .
  //        5 .     ,     .
  AmountWebTicks = WebToTicks("https://rannforex.com/rannforex/api/ticks/" + StringSubstr(_Symbol, 0, StringLen(_Symbol) - 5) + "/", WebTicks);

  if (AmountWebTicks)
  {
    int Pos = 0;

    if (AmountLoadTicks)
    {
      const datetime LastTime = LoadTicks[AmountLoadTicks - 1].time;


      for (; (Pos < AmountWebTicks) && (WebTicks[Pos].time <= LastTime); Pos++)
        ;
    }

    AmountWebTicks = (Pos < AmountWebTicks) ? ArrayCopy(LoadTicks, WebTicks, ArraySize(LoadTicks), Pos) : 0;
    AmountLoadTicks += AmountWebTicks;

    ulong StartTime = GetMicrosecondCount();

    MqlTick SaveTicks[];
    ArrayCopy(SaveTicks, WebTicks, 0, Pos);

    const int Size = TicksToFile(SaveTicks, sFileName);

    StartTime = (GetMicrosecondCount() - StartTime) / 1000;

    Print((string)Size + " new ticks saved (" + (string)StartTime +
          " ms). HistoryTicks' amount = " + (string)AmountLoadTicks +
          " - " + sFileName);
  }
#endif // HISTORYTICKS_RANNFOREX

  const int InitAmountTicks = (InitHoursTicks > 0) ? AmountLoadTicks - GetPosTick(LoadTicks, GetTime(InitHoursTicks * 3600)) : 0;

  AmountCopyTicks = (AmountLoadTicks > InitAmountTicks) ? ArrayCopy(HistoryTicks, LoadTicks, 0, AmountLoadTicks - InitAmountTicks, InitAmountTicks)
                                                        : ArrayCopy(HistoryTicks, LoadTicks);

  AmountTicks = AmountCopyTicks;

  if (!(GlobalVariableTemp(Name) && //       HDD.
        SaveCounter() && GlobalVariableSet(Name, AmountTicks)))
    ALERT(GlobalVariableSet(Name, AmountTicks));

  Alert("From " + sFileName + " read " + (string)AmountTicks + "/" + (string)AmountLoadTicks + " old ticks" +
        (AmountTicks ? ": " + (string)HistoryTicks[0].time + " - " + (string)HistoryTicks[AmountTicks - 1].time : "."));

  MqlTick Tick = {0};
  const bool TickInit = SymbolInfoTick(_Symbol, Tick) && (_R(Tick) != PrevTick);

  PrevTick = Tick;

  if (!TickInit)
  {
    ALERT(TickInit);

    PrevTick.time_msc = 0;
  }

  if (AmountTicks && (!TickInit || (PrevTick.time_msc <= HistoryTicks[AmountTicks - 1].time_msc)))
  {
    PrevTick = HistoryTicks[AmountTicks - 1];

    PrevTick.flags = Tick.flags; //    .
  }

  return;
}

void SaveTicks( const bool Force = false )
{
  if (AmountTicksToSave && (Force || (AmountTicks > AmountCopyTicks + (int)AmountTicksToSave)))
  {
    ulong StartTime = GetMicrosecondCount();

    MqlTick SaveTicks[];
    ArrayCopy(SaveTicks, HistoryTicks, 0, AmountCopyTicks);

    const int Size = TicksToFile(SaveTicks, sFileName);

    StartTime = (GetMicrosecondCount() - StartTime) / 1000;

    if (Size)
      Print((string)Size + " new ticks saved (" + (string)StartTime +
            " ms). HistoryTicks' amount = " + (string)ArraySize(HistoryTicks) +
            " - " + sFileName);

    AmountCopyTicks = AmountTicks;
  }

  return;
}

void OnDeinit( const int Reason )
{
  if (Reason != REASON_INITFAILED)
  {
    if (GlobalVariableCheck(Name) && !GlobalVariableDel(Name))
      ALERT(GlobalVariableDel(Name));

    if (IsOnInit())
      SaveTicks(true);

    if (Reason != REASON_ACCOUNT)
    {
      EventKillTimer();

      Alert(Name + " - stopped!");
    }
  }
}

void SendHistoryTicks( void )
{
  static const RESOURCEDATA<MqlTick> Resource("::" + Name);
  static const string TransferName = Name + "_Transfer";

  if (GlobalVariableCheck(TransferName))
    Resource = HistoryTicks;
  else
    Resource.Free();

  return;
}

// https://www.mql5.com/ru/forum/1111/page2961#comment_20811791
bool SaveCounter( void )
{
  static const RESOURCE Resource("::" + Name + "_Counter");
  static uint Data[1];

  Data[0] = AmountTicks;

  Resource.Set(Data);

  return(true);
}

void OnTimer()
{
  if (IsOnInit())
  {
    SaveTicks();
    SendHistoryTicks();
  }

  return;
}

#define RESERVE_HISTORY 100000

#ifdef __MQL5__
int OnCalculate( const int, const int, const int, const double &[] )
#else // __MQL5__
int OnCalculate( const int, const int, const datetime &[], const double &[], const double &[],
                 const double &[], const double &[], const long &[], const long &[], const int &[] )
#endif // __MQL5__
{
  static MqlTick Tick = {0};
  static const long MyChart = ChartID();

  if (IsOnInit())
  {
    if (SymbolInfoTick(_Symbol, Tick) && (_R(Tick) != PrevTick) && (Tick.time_msc >= PrevTick.time_msc))
    {
      PrevTick = Tick;

      ARRAY::AddElement(HistoryTicks, Tick, RESERVE_HISTORY);
      AmountTicks++;

      if (!(SaveCounter() && GlobalVariableSet(Name, AmountTicks)))
        ALERT(GlobalVariableSet(Name, AmountTicks));

      const string Str = DATA_STRING::ToString(Tick);

      if (AllCharts)
      {
        long Chart = ::ChartFirst();

        while (Chart != -1)
        {
          if ((inSelfEvent || (Chart != MyChart)) &&
              (ChartSymbol(Chart) == _Symbol) && !EventChartCustom(Chart, 0, AmountTicks, 0, Str))
            ALERT(EventChartCustom(Chart, 0, AmountTicks, 0, Str));

          Chart = ::ChartNext(Chart);
        }
      }
      else if (inSelfEvent && !EventChartCustom(MyChart, 0, AmountTicks, 0, Str))
        ALERT(EventChartCustom(MyChart, 0, AmountTicks, 0, Str));
    }

    SendHistoryTicks();
  }

  return(0);
}

#undef RESERVE_HISTORY
#undef ALERT